# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****

"""
This is a sample plugin for the metlog clien

You can find an entrypoint to it in the metlog-py setup.py file in the
entry_points section.

Each plugin is bound into the metlog client if and only if there is
configuration for the plugin.
"""
METLOG_PLUGIN_NAME = 'exception'

def config_plugin(config_dict):
    # Normally, the config_dict is unwrapped to get
    # some variables that can be used in the closure
    # to disable features
    default_verbose = config_dict.pop('verbose', False)

    import copy
    mycopy = copy.deepcopy(config_dict)

    def my_plugin(self, *args, **kwargs):
        # Most real plugin methods will use the variables captured in the
        # config_dict to override arguments passed in by the developer through
        # *args and **kwargs.  This allows operations to disable specific
        # features of metlog without having to deploy new code.

        if not (default_verbose and kwargs.get('verbose', False)):
            # Just skip early if any verbose flag has been set to False
            return
        return mycopy

    my_plugin.metlog_name = METLOG_PLUGIN_NAME

    return my_plugin



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#
# ***** END LICENSE BLOCK *****
from __future__ import absolute_import
from metlog.client import MetlogClient
from metlog.logging import hook_logger
from mock import Mock
from nose.tools import eq_

import logging


class TestLoggingHook(object):
    logger = 'tests'

    def setUp(self):
        self.mock_sender = Mock()
        self.client = MetlogClient(self.mock_sender, self.logger)

    def tearDown(self):
        del self.mock_sender

    def test_logging_handler(self):
        logger = logging.getLogger('demo')
        hook_logger('demo', self.client)
        msg = "this is an info message"
        logger.info(msg)
        eq_(msg, self.mock_sender.send_message.call_args[0][0]['payload'])



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#
# ***** END LICENSE BLOCK *****
from metlog.client import MetlogClient
from metlog.client import SEVERITY
from metlog.senders import DebugCaptureSender
from nose.tools import eq_, ok_
import json
import random
import threading


class TestMetlogClientFilters(object):
    logger = 'tests'
    timer_name = 'test'

    def setUp(self):
        self.sender = DebugCaptureSender()
        self.client = MetlogClient(self.sender, self.logger)

    def tearDown(self):
        del self.sender
        del self.client

    def test_severity_max(self):
        from metlog.filters import severity_max_provider
        self.client.filters = [severity_max_provider(severity=SEVERITY.ERROR)]
        payload = 'foo'
        self.client.debug(payload)
        self.client.info(payload)
        self.client.warn(payload)
        self.client.error(payload)
        self.client.exception(payload)
        self.client.critical(payload)
        # only half of the messages should have gone out
        eq_(len(self.sender.msgs), 3)
        # make sure it's the right half
        for json_msg in self.sender.msgs:
            msg = json.loads(json_msg)
            ok_(msg['severity'] <= SEVERITY.ERROR)

    def test_type_blacklist(self):
        from metlog.filters import type_blacklist_provider
        type_blacklist = type_blacklist_provider(types=set(['foo']))
        self.client.filters = [type_blacklist]
        choices = ['foo', 'bar']
        notfoos = 0
        for i in range(10):
            choice = random.choice(choices)
            if choice != 'foo':
                notfoos += 1
            self.client.metlog(choice, payload='msg')
        eq_(len(self.sender.msgs), notfoos)

    def test_type_whitelist(self):
        from metlog.filters import type_whitelist_provider
        type_whitelist = type_whitelist_provider(types=set(['foo']))
        self.client.filters = [type_whitelist]
        choices = ['foo', 'bar']
        foos = 0
        for i in range(10):
            choice = random.choice(choices)
            if choice == 'foo':
                foos += 1
            self.client.metlog(choice, payload='msg')
        eq_(len(self.sender.msgs), foos)

    def test_type_severity_max(self):
        from metlog.filters import type_severity_max_provider
        config = {'types': {'foo': {'severity': 3},
                            'bar': {'severity': 5},
                            },
                  }
        type_severity_max = type_severity_max_provider(**config)
        self.client.filters = [type_severity_max]
        for msgtype in ['foo', 'bar']:
            for sev in range(8):
                self.client.metlog(msgtype, severity=sev, payload='msg')
        eq_(len(self.sender.msgs), 10)
        msgs = [json.loads(msg) for msg in self.sender.msgs]
        foos = [msg for msg in msgs if msg['type'] == 'foo']
        eq_(len(foos), 4)
        bars = [msg for msg in msgs if msg['type'] == 'bar']
        eq_(len(bars), 6)



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#
# ***** END LICENSE BLOCK *****
"""
Compatibility layer that allows metlog code to generate output using Python's
standard library's `logging` module.
"""
from __future__ import absolute_import
from metlog.client import SEVERITY
import logging
try:
    import simplesjson as json
except ImportError:
    import json  # NOQA

# maps metlog message 'severity' to logging message 'level'
SEVERITY_MAP = {
    SEVERITY.EMERGENCY: logging.CRITICAL,
    SEVERITY.ALERT: logging.CRITICAL,
    SEVERITY.CRITICAL: logging.CRITICAL,
    SEVERITY.ERROR: logging.ERROR,
    SEVERITY.WARNING: logging.WARN,
    SEVERITY.NOTICE: logging.INFO,
    SEVERITY.INFORMATIONAL: logging.INFO,
    SEVERITY.DEBUG: logging.DEBUG,
    }


class StdLibLoggingSender(object):
    """
    Sender that passes messages off to Python stdlib's `logging` module for
    delivery. Messages will be handled one of three ways:

    1. Message payload extracted and sent on as plain text.
    2. Message serialized to JSON and sent on in its entirety.
    3. Message dropped.

    Sender is configurable to allow specification of which message types should
    be handled in which manner.
    """
    def __init__(self, logger_name=None, payload_types=None, json_types=None):
        """
        :param logger_name: Name of logger that should be fetched from logging
                            module.
        :param payload_types: Sequence of message types that should have their
                              payloads extracted and sent on as text.
        :param json_types: Sequence of message types that should be serialized
                           to JSON and sent on.
        """
        if logger_name is None:
            self.logger = logging.getLogger()
        else:
            self.logger = logging.getLogger(logger_name)
        if payload_types is None:
            payload_types = ['oldstyle']
        if isinstance(payload_types, basestring):
            payload_types = [payload_types]
        self.payload_types = set(payload_types)
        if json_types is None:
            json_types = ['*']
        if isinstance(json_types, basestring):
            json_types = [json_types]
        self.json_types = set(json_types)

    def send_message(self, msg):
        if msg['type'] in self.payload_types or '*' in self.payload_types:
            logging_msg = msg['payload']
        elif msg['type'] in self.json_types or '*' in self.json_types:
            logging_msg = json.dumps(msg)
        else:
            # drop it
            return
        lvl = SEVERITY_MAP.get(msg['severity'], logging.WARN)
        self.logger.log(lvl, logging_msg)



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#
# ***** END LICENSE BLOCK *****
from metlog.exceptions import EnvironmentNotFoundError
from metlog.client import MetlogClient
from metlog.config import client_from_text_config
from metlog.senders import DebugCaptureSender
from mock import Mock
from mock import patch
from nose.tools import assert_raises, eq_, ok_
import os


MockSender = Mock()


def test_simple_config():
    cfg_txt = """
    [metlog_config]
    sender_class = metlog.senders.DebugCaptureSender
    """
    client = client_from_text_config(cfg_txt, 'metlog_config')
    eq_(client.__class__, MetlogClient)
    eq_(client.sender.__class__, DebugCaptureSender)


def test_multiline_config():
    cfg_txt = """
    [metlog_config]
    sender_class = metlog.tests.test_config.MockSender
    sender_multi = foo
                   bar
    """
    client = client_from_text_config(cfg_txt, 'metlog_config')
    ok_(isinstance(client.sender, Mock))
    MockSender.assert_called_with(multi=['foo', 'bar'])


def test_environ_vars():
    env_var = 'SENDER_TEST'
    marker = object()
    orig_value = marker
    if env_var in os.environ:
        orig_value = os.environ[env_var]
    os.environ[env_var] = 'metlog.senders.DebugCaptureSender'
    cfg_txt = """
    [test1]
    sender_class = ${SENDER_TEST}
    """
    client = client_from_text_config(cfg_txt, 'test1')
    eq_(client.sender.__class__, DebugCaptureSender)

    cfg_txt = """
    [test1]
    sender_class = ${NO_SUCH_VAR}
    """
    assert_raises(EnvironmentNotFoundError, client_from_text_config,
                  cfg_txt, 'test1')
    if orig_value is not marker:
        os.environ[env_var] = orig_value
    else:
        del os.environ[env_var]


def test_int_bool_conversions():
    cfg_txt = """
    [metlog_config]
    sender_class = metlog.tests.test_config.MockSender
    sender_integer = 123
    sender_true1 = True
    sender_true2 = t
    sender_true3 = Yes
    sender_true4 = on
    sender_false1 = false
    sender_false2 = F
    sender_false3 = no
    sender_false4 = OFF
    """
    client = client_from_text_config(cfg_txt, 'metlog_config')
    ok_(isinstance(client.sender, Mock))
    MockSender.assert_called_with(integer=123, true1=True, true2=True,
                                  true3=True, true4=True, false1=False,
                                  false2=False, false3=False, false4=False)


def test_global_config():
    cfg_txt = """
    [metlog]
    sender_class = metlog.senders.DebugCaptureSender
    global_foo = bar
    global_multi = one
                   two
    """
    client_from_text_config(cfg_txt, 'metlog')
    from metlog.holder import CLIENT_HOLDER
    expected = {'foo': 'bar', 'multi': ['one', 'two']}
    eq_(expected, CLIENT_HOLDER.global_config)


def test_filters_config():
    cfg_txt = """
    [metlog]
    sender_class = metlog.senders.DebugCaptureSender
    [metlog_filter_sev_max]
    provider = metlog.filters.severity_max_provider
    severity = 6
    [metlog_filter_type_whitelist]
    provider = metlog.filters.type_whitelist_provider
    types = foo
            bar
            baz
    """
    client = client_from_text_config(cfg_txt, 'metlog')
    eq_(len(client.filters), 2)

    severity_max = client.filters[0]
    eq_(severity_max.func_name, 'severity_max')
    msg = {'severity': 6}
    ok_(severity_max(msg))
    msg = {'severity': 7}
    ok_(not severity_max(msg))

    type_whitelist = client.filters[1]
    eq_(type_whitelist.func_name, 'type_whitelist')
    msg = {'type': 'bar'}
    ok_(type_whitelist(msg))
    msg = {'type': 'bawlp'}
    ok_(not type_whitelist(msg))


def test_plugins_config():
    cfg_txt = """
    [metlog]
    sender_class = metlog.senders.DebugCaptureSender
    [metlog_plugin_dummy]
    provider=metlog.tests.plugin:config_plugin
    verbose=True
    foo=bar
    some_list = dog
                cat
                bus
    port=8080
    host=lolcathost
    """
    client = client_from_text_config(cfg_txt, 'metlog')
    actual = client.dummy(verbose=True)
    expected = {'host': 'lolcathost',
     'foo': 'bar', 'some_list': ['dog', 'cat', 'bus'],
     'port': 8080}
    assert actual == expected


def test_handshake_sender_no_backend():
    cfg_txt = """
    [metlog]
    sender_class = metlog.senders.ZmqHandshakePubSender
    sender_handshake_bind = tcp://localhost:5180
    sender_connect_bind = tcp://localhost:5190
    sender_handshake_timeout = 200
    sender_hwm = 100
    """
    import json
    import sys   # NOQA
    msg = {'milk': 'shake'}
    expected = "%s\n" % json.dumps(msg)
    with patch('sys.stderr') as mock_stderr:
        with patch('metlog.senders.zmq.Pool.start_reconnecting'):
            client = client_from_text_config(cfg_txt, 'metlog')
            client.send_message(msg)
            eq_(mock_stderr.write.call_count, 1)
            eq_(mock_stderr.flush.call_count, 1)
            call_args = mock_stderr.write.call_args[0]
            eq_(call_args[0], expected)


def test_handshake_sender_with_backend():
    cfg_txt = """
    [metlog]
    sender_class = metlog.senders.ZmqHandshakePubSender
    sender_handshake_bind = tcp://localhost:5180
    sender_connect_bind = tcp://localhost:5190
    sender_handshake_timeout = 200
    sender_hwm = 100
    sender_livecheck = 30
    """
    import json
    import sys   # NOQA

    # Redirect stderr
    with patch('sys.stderr') as mock_stderr:

        # Patch the reconnect_clients call so that we don't spawn a
        # background thread to bind to a server
        with patch('metlog.senders.zmq.Pool.start_reconnecting'):

            client = client_from_text_config(cfg_txt, 'metlog')

            # Now patch the ZmqHandshakePubSender and replace the pool
            # with a mock - this will make sure that all requests to
            # obtain a new 0mq socket will just pass so nothing will
            # go to stderr
            eq_(client.sender.pool._livecheck, 30)
            with patch.object(client.sender, 'pool') as mock_pool:
                msg = {'milk': 'shake'}

                # Note that this JSON dump does *not* have a newline appended
                # to it.  Only JSON messages to stderr have the newline
                expected = json.dumps(msg)

                client.send_message(msg)
                eq_(mock_stderr.write.call_count, 0)
                eq_(mock_stderr.flush.call_count, 0)

            # Check that we called send once with the proper JSON
            # string to the pool
            eq_(mock_pool.send.call_count, 1)
            call_args = mock_pool.send.call_args[0]
            eq_(call_args[0], expected)

def test_plugin_override():
    cfg_txt = """
    [metlog]
    sender_class = metlog.senders.DebugCaptureSender

    [metlog_plugin_exception]
    override=True
    provider=metlog.tests.plugin:config_plugin
    """
    client = client_from_text_config(cfg_txt, 'metlog')
    eq_('dummy', client.dummy.metlog_name)

    cfg_txt = """
    [metlog]
    sender_class = metlog.senders.DebugCaptureSender
    [metlog_plugin_exception]
    provider=metlog.tests.plugin_exception:config_plugin
    """
    # Failure to set an override argument will throw an exception
    assert_raises(SyntaxError, client_from_text_config, cfg_txt, 'metlog')



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#   Victor Ng (vng@mozilla.com)
#
# ***** END LICENSE BLOCK *****
from metlog.client import SEVERITY
from metlog.senders.udp import UdpSender
from metlog.senders.dev import StdOutSender
from metlog.senders.logging import StdLibLoggingSender
from metlog.senders.zmq import ZmqPubSender, zmq
from mock import patch
from nose.plugins.skip import SkipTest
from nose.tools import eq_

import json
import logging
import threading
import time


class TestZmqPubSender(object):
    logger = 'tests'

    def setUp(self):
        if zmq is None:
            raise(SkipTest)
        self.context_patcher = patch.object(ZmqPubSender, '_zmq_context')
        self.mock_zmq_context = self.context_patcher.start()
        self.sender = self._make_one()

    def tearDown(self):
        self.mock_zmq_context.stop()

    def _make_one(self):
        return ZmqPubSender(bindstrs='bindstr', pool_size=2)

    def test_publ_threadsafe(self):

        def reentrant():
            self.sender.send_message('foo')

        t0 = threading.Thread(target=reentrant)
        t1 = threading.Thread(target=reentrant)
        t0.start()
        time.sleep(0.01)  # give it time to ensure publisher is accessed
        t1.start()
        time.sleep(0.01)  # give it time to ensure publisher is accessed
        # the socket call should have happened twice, once for each thread
        mock_socket = self.mock_zmq_context.socket()
        eq_(mock_socket.connect.call_count, 2)
        eq_(mock_socket.send.call_count, 2)

    def test_send(self):
        msg = {'this': 'is',
               'a': 'test',
               'payload': 'PAYLOAD'}
        json_msg = json.dumps(msg)
        self.sender.send_message(msg)
        mock_socket = self.mock_zmq_context.socket()

        eq_(mock_socket.connect.call_count, 2)
        eq_(mock_socket.connect.call_args, (('bindstr',), {}))
        eq_(mock_socket.send.call_count, 1)

        mock_socket.send.assert_called_with(json_msg)

    def test_debug_stderr(self):
        msg = {'milk': 'shake'}
        json_msg = json.dumps(msg)
        self.sender.debug_stderr = True
        with patch('sys.stderr') as mock_stderr:
            self.sender.send_message(msg)
            eq_(mock_stderr.write.call_count, 1)
            eq_(mock_stderr.flush.call_count, 1)
            mock_stderr.write.assert_called_with(json_msg + '\n')


def formatter(msg):
    output = []
    for key, value in msg.items():
        output.append('%s;;;%s' % (str(key), str(value)))
    return '\n'.join(output)


@patch('sys.stdout')
class TestStdOutSender(object):
    def _make_one(self, formatter=None):
        return StdOutSender(formatter=formatter)

    def setUp(self):
        self.msg = {'this': 'is',
                    'a': 'test',
                    'payload': 'PAYLOAD'}

    def test_default_formatter(self, mock_stdout):
        sender = self._make_one()
        sender.send_message(self.msg)
        eq_(mock_stdout.write.call_count, 1)
        eq_(mock_stdout.flush.call_count, 1)
        write_args = mock_stdout.write.call_args
        eq_(json.loads(write_args[0][0]), self.msg)

    def test_custom_formatter(self, mock_stdout):
        sender = self._make_one(formatter=formatter)
        sender.send_message(self.msg)
        eq_(mock_stdout.write.call_count, 1)
        eq_(mock_stdout.flush.call_count, 1)
        mock_stdout.write.assert_called_with(formatter(self.msg) + '\n')

    def test_custom_formatter_dotted(self, mock_stdout):
        dotted = 'metlog.tests.test_senders.formatter'
        sender = self._make_one(formatter=dotted)
        sender.send_message(self.msg)
        eq_(mock_stdout.write.call_count, 1)
        eq_(mock_stdout.flush.call_count, 1)
        mock_stdout.write.assert_called_with(formatter(self.msg) + '\n')


@patch('metlog.senders.logging.logging')
class TestLoggingSender(object):
    msgs = [{'type': 'oldstyle', 'payload': 'oldstyle',
             'severity': SEVERITY.WARNING},
            {'type': 'this', 'payload': 'this', 'severity': SEVERITY.ERROR},
            {'type': 'that', 'payload': 'that',
             'severity': SEVERITY.INFORMATIONAL},
            {'type': 'the other', 'payload': 'the other',
             'severity': SEVERITY.DEBUG},
            ]

    def _make_one(self, *args, **kwargs):
        return StdLibLoggingSender(*args, **kwargs)

    def _send_em(self, sender):
        for msg in self.msgs:
            sender.send_message(msg)

    def test_defaults(self, mock_logging):
        sender = self._make_one()
        self._send_em(sender)
        log = mock_logging.getLogger().log
        eq_(log.call_count, 4)
        log.assert_any_call(logging.WARN, 'oldstyle')
        log.assert_any_call(logging.ERROR, json.dumps(self.msgs[1]))
        log.assert_any_call(logging.INFO, json.dumps(self.msgs[2]))
        log.assert_called_with(logging.DEBUG, json.dumps(self.msgs[3]))

    def test_alternate_logger_name(self, mock_logging):
        name = 'logger_name'
        sender = self._make_one(name)
        self._send_em(sender)
        log = mock_logging.getLogger(name).log
        eq_(log.call_count, 4)
        log.assert_any_call(logging.WARN, 'oldstyle')
        log.assert_any_call(logging.ERROR, json.dumps(self.msgs[1]))
        log.assert_any_call(logging.INFO, json.dumps(self.msgs[2]))
        log.assert_called_with(logging.DEBUG, json.dumps(self.msgs[3]))

    def test_specific_types(self, mock_logging):
        sender = self._make_one(payload_types=['this', 'that'],
                                json_types=['the other'])
        self._send_em(sender)
        log = mock_logging.getLogger().log
        eq_(log.call_count, 3)
        log.assert_any_call(logging.ERROR, 'this')
        log.assert_any_call(logging.INFO, 'that')
        log.assert_called_with(logging.DEBUG, json.dumps(self.msgs[3]))

    def test_payload_all(self, mock_logging):
        sender = self._make_one(payload_types=['*'])
        self._send_em(sender)
        log = mock_logging.getLogger().log
        eq_(log.call_count, 4)
        log.assert_any_call(logging.WARN, 'oldstyle')
        log.assert_any_call(logging.ERROR, 'this')
        log.assert_any_call(logging.INFO, 'that')
        log.assert_any_call(logging.DEBUG, 'the other')


class TestUdpSender(object):
    def _make_one(self, host='127.0.0.1', port=5565):
        return UdpSender(host=host, port=port)

    def setUp(self):
        self.sender = self._make_one()
        self.socket_patcher = patch.object(self.sender, 'socket')
        self.mock_socket = self.socket_patcher.start()
        self.msg = {'this': 'is',
                    'a': 'test',
                    'payload': 'PAYLOAD'}

    def test_sender(self):
        self.sender.send_message(self.msg)
        eq_(self.mock_socket.sendto.call_count, 1)
        write_args = self.mock_socket.sendto.call_args
        assert write_args[0][1] == ('127.0.0.1', 5565)
        assert json.loads(write_args[0][0]) == self.msg



#
# Synchronized publisher. This is like a metlog cliennt
#
import zmq
import Queue
import sys
from contextlib import contextmanager
import threading
import time

class Client(object):
    def __init__(self, context, handshake_bind, connect_bind,
            handshake_timeout=200):

        self.context = context

        self.handshake_bind = handshake_bind
        self.connect_bind = connect_bind

        self.handshake_timeout = handshake_timeout

        self.handshake_socket = None
        self.socket = None

        self._connected = False

    def connect(self):
        # Socket to send handshake signals
        self.handshake_socket = self.context.socket(zmq.REQ)
        self.handshake_socket.connect(self.handshake_bind)
        self.handshake_socket.setsockopt(zmq.LINGER, 0)

        # Socket to actually doi pub/sub
        self.socket = self.context.socket(zmq.PUB)
        self.socket.connect(self.connect_bind)
        self.socket.setsockopt(zmq.LINGER, 0)

        poll = zmq.Poller()
        poll.register(self.handshake_socket, zmq.POLLIN)

        self.handshake_socket.send("")
        socks = dict(poll.poll(self.handshake_timeout))

        if socks.get(self.handshake_socket) == zmq.POLLIN:
            self.handshake_socket.recv()
            self.handshake_socket.close()
            self._connected = True
            return True
        else:
            self._connected = False
            self.close()
            return False

    def send(self, msg):
        if self._connected:
            self.socket.send(msg)
        elif self.connect():
            self.socket.send(msg)
        else:
            sys.stderr.write(msg)
            sys.stderr.flush()

    def close(self):
        try:
            self.handshake_socket.close()
        except ZeroMQError, err:
            pass

        try:
            self.socket.close()
        except ZeroMQError, err:
            pass

def client_factory(context):
    def get_client():
        return Client(context, 'tcp://localhost:5562', 'tcp://localhost:5561')
    return get_client

class Pool(object):
    """
    This is a threadsafe poool of 0mq clients.

    :param client_factory:
        a factory function which takes no arguments and will generate
        client instances. Clients are responsible to rebind themselves
        if necessary
    :param size: The number of clients to create in the pool
    """

    def __init__(self, client_factory, size=10):
        self._clients = Queue.Queue()
        for i in range(size):
            self._clients.put(client_factory())

    def send(self, msg):
        sock = None
        try:
            sock = self.socket()
            sock.send(msg)
        except Queue.Empty:
            # Sometimes, we'll get nothing
            sys.stderr.write(msg)
        finally:
            if sock:
                self._clients.put(sock)

    def socket(self):
        return self._clients.get()

def main():
    context = zmq.Context()

    pool = Pool(client_factory(context))

    def threaded_send():
        for i in range(10000):
            pool.send('Rhubarb');

    for i in range(10):
        t = threading.Thread(target=threaded_send)
        t.daemon = True
        t.start()

    time.sleep(10)
    pool.send('END')

if __name__ == '__main__':
    main()




# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at http://mozilla.org/MPL/2.0/.
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#   Victor Ng (vng@mozilla.com)
#
# ***** END LICENSE BLOCK *****
from __future__ import absolute_import

from metlog.client import SEVERITY
import logging


class MetlogHandler(logging.Handler):
    def __init__(self, metlog_client):
        logging.Handler.__init__(self)
        self.metlog_client = metlog_client

    def emit(self, record):
        if record.levelno == logging.DEBUG:
            severity = SEVERITY.DEBUG
        elif record.levelno == logging.INFO:
            severity = SEVERITY.INFORMATIONAL
        elif record.levelno == logging.WARNING:
            severity = SEVERITY.WARNING
        elif record.levelno == logging.ERROR:
            severity = SEVERITY.ERROR
        else:  # critical
            severity = SEVERITY.CRITICAL

        self.metlog_client.metlog(type='oldstyle', severity=severity,
                                  payload=record.msg)


def hook_logger(logger_name, client):
    """
    Used to hook metlog into the Python stdlib logging framework. Registers a
    logging module handler that delegates to a MetlogClient for actual message
    delivery.

    :param name: Name of the stdlib logging `logger` object for which the
                 handler should be registered.
    :param client: MetlogClient instance that the registered handler will use
                   for actual message delivery.
    """
    logger = logging.getLogger(logger_name)
    # first check to see if we're already registered
    for existing in logger.handlers:
        if (isinstance(existing, MetlogHandler) and
            existing.metlog_client is client):
            # already done, do nothing
            return
    logger.addHandler(MetlogHandler(client))




#
# Synchronized subscriber. This is like logstash
#
import zmq
import threading

context = zmq.Context()

def sync_thread():
    # Second, synchronize with publisher
    syncclient = context.socket(zmq.REP)
    syncclient.bind('tcp://*:5562')

    while True:
        # wait for a synchronization request
        syncclient.recv()

        # send synchronization reply
        syncclient.send('')

def main():

    # First, connect our subscriber socket
    subscriber = context.socket(zmq.SUB)
    subscriber.bind('tcp://*:5561')
    subscriber.setsockopt(zmq.SUBSCRIBE, "")

    t = threading.Thread(target=sync_thread)
    t.daemon = True
    t.start()


    # Third, get our updates and report how many we got
    nbr = 0
    while True:
        msg = subscriber.recv()
        if msg == 'END':
            break
        nbr += 1

    print 'Received %d updates' % nbr

if __name__ == '__main__':
    main()




# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#   James Socol (james@mozilla.com)
#   Victor Ng (vng@mozilla.com)
#
# ***** END LICENSE BLOCK *****
from __future__ import absolute_import
import os
import random
import socket
import sys
import threading
import time
import traceback
import types

from datetime import datetime
from functools import wraps
from metlog.senders import NoSendSender


class SEVERITY:
    '''
    Put a namespace around RFC 3164 syslog messages
    '''
    EMERGENCY = 0
    ALERT = 1
    CRITICAL = 2
    ERROR = 3
    WARNING = 4
    NOTICE = 5
    INFORMATIONAL = 6
    DEBUG = 7


class _NoOpTimer(object):
    """
    A bogus timer object that will act as a contextdecorator but which
    doesn't actually do anything.
    """
    def __init__(self):
        self.start = None
        self.result = None

    def __call__(self, fn):
        return fn

    def __enter__(self):
        return self

    def __exit__(self, typ, value, tb):
        return False


class _Timer(object):
    """A contextdecorator for timing."""

    def __init__(self, client, name, msg_data):
        # most attributes on a _Timer object should be threadlocal, except for
        # a few which we put directly in the __dict__
        self.__dict__['client'] = client
        self.__dict__['_local'] = threading.local()
        self.__dict__['name'] = name
        self.msg_data = msg_data

    def __delattr__(self, attr):
        """Store thread-local data safely."""
        delattr(self._local, attr)

    def __getattr__(self, attr):
        """Store thread-local data safely."""
        return getattr(self._local, attr)

    def __setattr__(self, attr, value):
        """Store thread-local data safely."""
        setattr(self._local, attr, value)

    def __call__(self, fn):
        """
        Support for use as a decorator.
        """
        if not callable(fn):
            # whoops, can't decorate if we're not callable
            raise ValueError('Timer objects can only wrap callable objects.')

        @wraps(fn)
        def wrapped(*a, **kw):
            with self:
                return fn(*a, **kw)
        return wrapped

    def __enter__(self):
        self.start = time.time()
        self.result = None
        return self

    def __exit__(self, typ, value, tb):
        elapsed = time.time() - self.start
        elapsed = int(round(elapsed * 1000))  # Convert to ms.
        self.result = elapsed
        self.client.timer_send(self.name, elapsed, **self.msg_data)
        return False


class MetlogClient(object):
    """
    Client class encapsulating metlog API, and providing storage for default
    values for various metlog call settings.
    """
    # envelope version, only changes when the message format changes
    env_version = '0.8'

    def __init__(self, sender, logger, severity=6,
                 disabled_timers=None, filters=None):
        """
        :param sender: A sender object used for actual message delivery.
        :param logger: Default `logger` value for all sent messages.
                       This is commonly set to be the name of the
                       current application and is not modified for
                       different instances of metlog within the
                       scope of the same application.
        :param severity: Default `severity` value for all sent messages.
        :param disabled_timers: Sequence of string tokens identifying timers
                                that should be deactivated.
        :param filters: A sequence of filter callables.
        """
        self.setup(sender, logger, severity, disabled_timers, filters)
        self._dynamic_methods = {}
        self._timer_obs = {}
        self._noop_timer = _NoOpTimer()
        self.hostname = socket.gethostname()
        self.pid = os.getpid()

        # seed random for rate calculations
        random.seed()

    def setup(self, sender=None, logger='', severity=6, disabled_timers=None,
              filters=None):
        """
        :param sender: A sender object used for actual message delivery.
        :param logger: Default `logger` value for all sent messages.
        :param severity: Default `severity` value for all sent messages.
        :param disabled_timers: Sequence of string tokens identifying timers
                                that should be deactivated.
        :param filters: A sequence of filter callables.
        """
        if sender is None:
            sender = NoSendSender()
        self.sender = sender
        self.logger = logger
        self.severity = severity

        if disabled_timers is None:
            self._disabled_timers = set()
        else:
            self._disabled_timers = set(disabled_timers)
        if filters is None:
            filters = list()
        self.filters = filters

    @property
    def is_active(self):
        """
        Is this client ready to transmit messages? For now we assume that if
        the default sender (i.e. `NoSendSender`) has been replaced then we're
        good to go.
        """
        return not isinstance(self.sender, NoSendSender)

    def send_message(self, msg):
        """
        Apply any filters and, if required, pass message along to the sender
        for delivery.
        """
        for filter_fn in self.filters:
            if not filter_fn(msg):
                return
        self.sender.send_message(msg)

    def add_method(self, method, override=False):
        """
        Add a custom method to the MetlogClient instance.

        :param method: Callable that will be used as the method.
        :param override: Set this to the method name you want to
                         override. False indicates no override will
                         occur.
        """
        assert isinstance(method, types.FunctionType)

        # Obtain the metlog name directly from the method
        name = method.metlog_name
        if isinstance(override, basestring):
            name = override

        if override is False and hasattr(self, name):
            msg = "The name [%s] is already in use" % name
            raise SyntaxError(msg)

        self._dynamic_methods[name] = method
        meth = types.MethodType(method, self, self.__class__)
        setattr(self, name, meth)

    def metlog(self, type, timestamp=None, logger=None, severity=None,
               payload='', fields=None):
        """
        Create a single message and pass it to the sender for delivery.

        :param type: String token identifying the type of message payload.
        :param timestamp: Time at which the message is generated.
        :param logger: String token identifying the message generator.
        :param severity: Numerical code (0-7) for msg severity, per RFC 5424.
        :param payload: Actual message contents.
        :param fields: Arbitrary key/value pairs for add'l metadata.
        """
        timestamp = timestamp if timestamp is not None else datetime.utcnow()
        logger = logger if logger is not None else self.logger
        severity = severity if severity is not None else self.severity
        fields = fields if fields is not None else dict()

        if hasattr(timestamp, 'isoformat'):
            timestamp = timestamp.isoformat()
        full_msg = dict(type=type, timestamp=timestamp, logger=logger,
                        severity=severity, payload=payload, fields=fields,
                        env_version=self.env_version,
                        metlog_pid=self.pid,
                        metlog_hostname=self.hostname)
        self.send_message(full_msg)

    def timer(self, name, timestamp=None, logger=None, severity=None,
              fields=None, rate=1.0):
        """
        Return a timer object that can be used as a context manager or a
        decorator, generating a metlog 'timer' message upon exit.

        :param name: Required string label for the timer.
        :param timestamp: Time at which the message is generated.
        :param logger: String token identifying the message generator.
        :param severity: Numerical code (0-7) for msg severity, per RFC 5424.
        :param fields: Arbitrary key/value pairs for add'l metadata.
        :param rate: Sample rate, btn 0 & 1, inclusive (i.e. .5 = 50%). Sample
                     rate is enforced in this method, i.e. if a sample rate is
                     used then some percentage of the timers will do nothing.
        """
        # check if timer(s) is(are) disabled or if we exclude for sample rate
        if ((self._disabled_timers.intersection(set(['*', name]))) or
            (rate < 1.0 and random.random() >= rate)):
            return self._noop_timer
        msg_data = dict(timestamp=timestamp, logger=logger, severity=severity,
                        fields=fields, rate=rate)
        if name in self._timer_obs:
            timer = self._timer_obs[name]
            timer.msg_data = msg_data
        else:
            timer = _Timer(self, name, msg_data)
            self._timer_obs[name] = timer
        return timer

    def timer_send(self, name, elapsed, timestamp=None, logger=None,
                   severity=None, fields=None, rate=1.0):
        """
        Converts timing data into a metlog message for delivery.

        :param name: Required string label for the timer.
        :param elapsed: Elapsed time of the timed event, in ms.
        :param timestamp: Time at which the message is generated.
        :param logger: String token identifying the message generator.
        :param severity: Numerical code (0-7) for msg severity, per RFC 5424.
        :param fields: Arbitrary key/value pairs for add'l metadata.
        :param rate: Sample rate, btn 0 & 1, inclusive (i.e. .5 = 50%). Sample
                     rate is *NOT* enforced in this method, i.e. all messages
                     will be sent through to metlog, sample rate is purely
                     informational at this point.
        """
        payload = str(elapsed)
        fields = fields if fields is not None else dict()
        fields.update({'name': name, 'rate': rate})
        self.metlog('timer', timestamp, logger, severity, payload, fields)

    def incr(self, name, count=1, timestamp=None, logger=None, severity=None,
             fields=None, rate=1.0):
        """
        Sends an 'increment counter' message.

        :param name: String label for the counter.
        :param count: Integer amount by which to increment the counter.
        :param timestamp: Time at which the message is generated.
        :param logger: String token identifying the message generator.
        :param severity: Numerical code (0-7) for msg severity, per RFC 5424.
        :param fields: Arbitrary key/value pairs for add'l metadata.
        """
        if rate < 1 and random.random() >= rate:
            return
        payload = str(count)
        fields = fields if fields is not None else dict()
        fields['name'] = name
        fields['rate'] = rate
        self.metlog('counter', timestamp, logger, severity, payload, fields)

    # Standard Python logging API emulation
    def _oldstyle(self, severity, msg, *args, **kwargs):
        """Do any necessary string formatting and then generate the msg"""
        # if `args` is a mapping then extract it
        if (len(args) == 1 and hasattr(args[0], 'keys')
            and hasattr(args[0], '__getitem__')):
            args = args[0]
        if args:
            msg = msg % args
        exc_info = kwargs.get('exc_info', False)
        if exc_info:
            if not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()
            tb_lines = traceback.format_exception(exc_info[0], exc_info[1],
                                                  exc_info[2])
            s = ''.join(tb_lines)
            if s[-1:] == '\n':
                s = s[:-1]
            if msg[-1:] != '\n':
                msg = msg + '\n'
            try:
                msg = msg + s
            except UnicodeError:
                msg = msg + s.decode(sys.getfilesystemencoding())
        self.metlog(type='oldstyle', severity=severity, payload=msg)

    def debug(self, msg, *args, **kwargs):
        """ Log a DEBUG level message """
        self._oldstyle(SEVERITY.DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """ Log an INFO level message """
        self._oldstyle(SEVERITY.INFORMATIONAL, msg, *args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        """ Log a WARN level message """
        self._oldstyle(SEVERITY.WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """ Log an ERROR level message """
        self._oldstyle(SEVERITY.ERROR, msg, *args, **kwargs)

    def exception(self, msg, exc_info=True, *args, **kwargs):
        """ Log an ALERT level message """
        self._oldstyle(SEVERITY.ALERT, msg, exc_info=exc_info, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """ Log a CRITICAL level message """
        self._oldstyle(SEVERITY.CRITICAL, msg, *args, **kwargs)



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#   Victor Ng (vng@mozilla.com)
#
# ***** END LICENSE BLOCK *****
"""
This module provides helpers to handle MetlogClient configuration details.
"""
from metlog.client import MetlogClient
from metlog.exceptions import EnvironmentNotFoundError
from metlog.path import DottedNameResolver
from textwrap import dedent
import ConfigParser
import StringIO
import os
import re

_IS_INTEGER = re.compile('^-?[0-9].*')
_IS_ENV_VAR = re.compile('\$\{(\w.*)?\}')


def _get_env_val(match_obj):
    var = match_obj.groups()[0]
    if var not in os.environ:
        raise EnvironmentNotFoundError(var)
    return os.environ[var]


def _convert(value):
    """Converts a config value. Numeric integer strings are converted to
    integer values.  'True-ish' string values are converted to boolean True,
    'False-ish' to boolean False. Any alphanumeric (plus underscore) value
    enclosed within ${dollar_sign_curly_braces} is assumed to represent an
    environment variable, and will be converted to the corresponding value
    provided by os.environ.
    """
    def do_convert(value):
        if not isinstance(value, basestring):
            # we only convert strings
            return value

        value = value.strip()
        if _IS_INTEGER.match(value):
            try:
                return int(value)
            except ValueError:
                pass
        elif value.lower() in ('true', 't', 'on', 'yes'):
            return True
        elif value.lower() in ('false', 'f', 'off', 'no'):
            return False
        match_obj = _IS_ENV_VAR.match(value)
        if match_obj:
            return _get_env_val(match_obj)
        return value

    if isinstance(value, basestring) and '\n' in value:
        return [line for line in [do_convert(line)
                                  for line in value.split('\n')]
                if line.strip() != '']

    return do_convert(value)


def nest_prefixes(config_dict, prefixes=None, separator="_"):
    """
    Iterates through the `config_dict` keys, looking for any starting w/ one of
    a specific set of prefixes, moving those into a single nested dictionary
    keyed by the prefix value.

    :param config_dict: Dictionary to mutate. Will also be returned.
    :param prefixes: Sequence of prefixes to look for in `config_dict` keys.
    :param separator: String which separates prefix values from the rest of the
                      key.
    """
    if prefixes is None:
        prefixes = ['sender', 'global']
    for prefix in prefixes:
        prefix_dict = {}
        for key in config_dict.keys():
            full_prefix = prefix + separator
            if key.startswith(full_prefix):
                nested_key = key[len(full_prefix):]
                prefix_dict[nested_key] = config_dict[key]
        if prefix_dict:
            if prefix in config_dict:
                config_dict[prefix].update(prefix_dict)
            else:
                config_dict[prefix] = prefix_dict
    return config_dict


def client_from_dict_config(config, client=None, clear_global=False):
    """
    Configure a metlog client, fully configured w/ sender and plugins.

    :param config: Configuration dictionary.
    :param client: MetlogClient instance to configure. If None, one will be
                   created.
    :param clear_global: If True, delete any existing global config on the
                         CLIENT_HOLDER before applying new config.

    The configuration dict supports the following values:

    logger
      Metlog client default logger value.
    severity
      Metlog client default severity value.
    disabled_timers
      Sequence of string tokens identifying timers that are to be deactivated.
    filters
      Sequence of 2-tuples `(filter_provider, config)`. Each `filter_provider`
      is a dotted name referring to a function which, when called and passed
      the associated `config` dict as kwargs, will return a usable MetlogClient
      filter function.
    plugins
      Nested dictionary containing plugin configuration. Keys are the plugin
      names (i.e. the name the method will be given when attached to the
      client). Values are 2-tuples `(plugin_provider, config)`. Each
      `plugin_provider` is a dotted name referring to a function which, when
      called and passed the associated `config`, will return the usable plugin
      method.
    sender
      Nested dictionary containing sender configuration.
    global
      Dictionary to be applied to CLIENT_HOLDER's `global_config` storage.
      New config will overwrite any conflicting values, but will not delete
      other config entries. To delete, calling code should call the function
      with `clear_global` set to True.

    All of the configuration values are optional, but failure to include a
    sender may result in a non-functional Metlog client. Any unrecognized keys
    will be ignored.

    Note that any top level config values starting with `sender_` will be added
    to the `sender` config dictionary, overwriting any values that may already
    be set.

    The sender configuration supports the following values:

    class (required)
      Dotted name identifying the sender class to instantiate.
    args
      Sequence of non-keyword args to pass to sender constructor.
    <kwargs>
      All remaining key-value pairs in the sender config dict will be passed as
      keyword arguments to the sender constructor.
    """
    config = nest_prefixes(config)

    sender_config = config.get('sender', {})
    logger = config.get('logger', '')
    severity = config.get('severity', 6)
    disabled_timers = config.get('disabled_timers', [])
    filter_specs = config.get('filters', [])
    plugins_data = config.pop('plugins', {})
    global_conf = config.get('global', {})

    # update global config stored in CLIENT_HOLDER
    from metlog.holder import CLIENT_HOLDER
    if clear_global:
        CLIENT_HOLDER.global_config = {}
    CLIENT_HOLDER.global_config.update(global_conf)

    resolver = DottedNameResolver()

    # instantiate sender
    sender_clsname = sender_config.pop('class')
    sender_cls = resolver.resolve(sender_clsname)
    sender_args = sender_config.pop('args', tuple())
    sender = sender_cls(*sender_args, **sender_config)

    # initialize filters
    filters = [resolver.resolve(dotted_name)(**cfg)
               for (dotted_name, cfg) in filter_specs]

    # instantiate and/or configure client
    if client is None:
        client = MetlogClient(sender, logger, severity, disabled_timers,
                              filters)
    else:
        client.setup(sender, logger, severity, disabled_timers, filters)

    # initialize plugins and attach to client
    for section_name, plugin_spec in plugins_data.items():
        # each plugin spec is a 2-tuple: (dotted_name, cfg)
        plugin_config = plugin_spec[1]
        plugin_override = plugin_config.pop('override', False)
        plugin_fn = resolver.resolve(plugin_spec[0])(plugin_config)
        client.add_method(plugin_fn, plugin_override)

    return client


def dict_from_stream_config(stream, section):
    """
    Parses configuration from a stream and converts it to a dictionary suitable
    for passing to `client_from_dict_config`.

    :param stream: Stream object containing config information.
    :param section: INI file section containing the configuration we care
                    about.
    """
    config = ConfigParser.SafeConfigParser()
    config.readfp(stream)
    client_dict = {}

    # extract main client configuration
    for opt in config.options(section):
        client_dict[opt] = _convert(config.get(section, opt))

    # extract filter config from filter sections
    filters = []
    filter_sections = [n for n in config.sections()
                       if n.startswith('%s_filter' % section)]
    for filter_section in filter_sections:
        filter_config = {}
        for opt in config.options(filter_section):
            if opt == 'provider':
                # must be a dotted name string, don't convert
                dotted_name = config.get(filter_section, opt)
            else:
                filter_config[opt] = _convert(config.get(filter_section, opt))
        filters.append((dotted_name, filter_config))
    client_dict['filters'] = filters

    # extract plugin config from plugin sections
    plugins = {}
    plugin_sections = [n for n in config.sections()
                       if n.startswith("%s_plugin" % section)]
    for plugin_section in plugin_sections:
        plugin_name = plugin_section.replace("%s_plugin_" % section, '')
        plugin_config = {}
        provider = ''
        for opt in config.options(plugin_section):
            if opt == 'provider':
                # must be a dotted name string, don't convert
                provider = config.get(plugin_section, opt)
            else:
                plugin_config[opt] = _convert(config.get(plugin_section, opt))
        plugins[plugin_name] = (provider, plugin_config)
    client_dict['plugins'] = plugins

    client_dict = nest_prefixes(client_dict)
    return client_dict


def client_from_stream_config(stream, section, client=None,
                              clear_global=False):
    """
    Extract configuration data in INI format from a stream object (e.g. a file
    object) and use it to generate a Metlog client. Config values will be sent
    through the `_convert` function for possible type conversion.

    :param stream: Stream object containing config information.
    :param section: INI file section containing the configuration we care
                    about.
    :param client: MetlogClient instance to configure. If None, one will be
                   created.

    Note that all sender config options should be prefaced by "sender_", e.g.
    "sender_class" should specify the dotted name of the sender class to use.
    Similarly all extension method settings should be prefaced by
    "extensions_". Any values prefaced by "global_" will be added to the global
    config dictionary.
    """
    client_dict = dict_from_stream_config(stream, section)
    client = client_from_dict_config(client_dict, client)
    return client


def client_from_text_config(text, section, client=None, clear_global=False):
    """
    Extract configuration data in INI format from provided text and use it to
    configure a Metlog client. Text is converted to a stream and passed on to
    `client_from_stream_config`.

    :param text: INI text containing config information.
    :param section: INI file section containing the configuration we care
                    about.
    :param client: MetlogClient instance to configure. If None, one will be
                   created.
    """
    stream = StringIO.StringIO(dedent(text))
    return client_from_stream_config(stream, section, client)



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#
# ***** END LICENSE BLOCK *****
"""
Callables that provide message filters for the MetlogClient. Each provider
accepts needed filter configuration values, and returns a filter function
usable by a MetlogClient instance.

Each filter accepts a single `msg` dictionary argument and returns a boolean
value: True if a message *should* be delivered, False if a message *should not*
be delivered. Note that the `msg` dictionary *may* be mutated by the filter.
"""


def severity_max_provider(severity):
    """
    Filter if message severity is greater than specified `severity`.
    """
    def severity_max(msg):
        if msg['severity'] > severity:
            return False
        return True

    return severity_max


def type_blacklist_provider(types):
    """
    Filter if message type is in the `types` value.
    """
    def type_blacklist(msg):
        if msg['type'] in types:
            return False
        return True

    return type_blacklist


def type_whitelist_provider(types):
    """
    Filter if message type is NOT in the `types` value.
    """
    def type_whitelist(msg):
        if msg['type'] not in types:
            return False
        return True

    return type_whitelist


def type_severity_max_provider(types):
    """
    Filter if message type has specified maximum severity value and message
    severity is higher than this maximum. Each keyword argument key should be a
    message type name, and each keyword argument value should be the maximum
    allowed severity for that message type.
    """
    for msgtype in types:
        severity_filter = severity_max_provider(**types[msgtype])
        types[msgtype] = severity_filter

    def type_severity_max(msg):
        msgtype = msg['type']
        if msgtype not in types:
            return True
        severity_filter = types[msgtype]
        return severity_filter(msg)

    return type_severity_max



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#
# ***** END LICENSE BLOCK *****
try:
    import simplejson as json
except ImportError:
    import json  # NOQA
import sys

from metlog.path import resolve_name


class StreamSender(object):
    """
    Emits messages to a provided stream object.
    """
    def __init__(self, stream, formatter=None):
        """
        :param stream: Stream object to which the messages should be written.
        :param formatter: Optional callable (or dotted name identifier) that
                          accepts a msg dictionary and returns a formatted
                          string to be written to the stream.
        """
        self.stream = stream
        if formatter is None:
            self.formatter = self.default_formatter
        else:
            if not callable(formatter):
                formatter = resolve_name(formatter)
            self.formatter = formatter

    def default_formatter(self, msg):
        """
        Default formatter, just converts the message to 4-space-indented
        JSON.
        """
        return json.dumps(msg, indent=4)

    def send_message(self, msg):
        """Deliver message to the stream object."""
        output = self.formatter(msg)
        self.stream.write('%s\n' % output)
        self.stream.flush()


class StdOutSender(StreamSender):
    """
    Emits metlog messages to stdout.
    """
    def __init__(self, *args, **kwargs):
        super(StdOutSender, self).__init__(sys.stdout, *args, **kwargs)


class FileSender(StreamSender):
    """
    Emits messages to a filesystem file.
    """
    def __init__(self, filepath, *args, **kwargs):
        filestream = open(filepath, 'a')
        super(FileSender, self).__init__(filestream, *args, **kwargs)


class DebugCaptureSender(object):
    """
    Capture up to 100 metlog messages in a circular buffer for inspection
    later. This is only for DEBUGGING.  Do not use this for anything except
    development.
    """
    def __init__(self, **kwargs):
        import collections
        self.msgs = collections.deque(maxlen=100)
        for k, v in kwargs.items():
            # set arbitrary attributes, useful for testing
            setattr(self, k, v)

    def send_message(self, msg):
        """JSONify and append to the circular buffer."""
        json_msg = json.dumps(msg)
        self.msgs.append(json_msg)



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Victor Ng (vng@mozilla.com)
#
# ***** END LICENSE BLOCK *****

from __future__ import absolute_import

try:
    import simplejson as json
except:
    import json  # NOQA

import socket


class UdpSender(object):
    """
    Sends metlog messages out via a UDP socket.
    """

    def __init__(self, host, port):
        self._host = host
        self._port = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def send_message(self, msg):
        """
        Serialize and send a message off to the metlog listener.

        :param msg: Dictionary representing the message.
        """
        json_msg = json.dumps(msg)
        self.socket.sendto(json_msg, (self._host, self._port))



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#   Victor Ng (vng@mozilla.com)
#
# ***** END LICENSE BLOCK *****
from metlog.client import MetlogClient
from metlog.config import client_from_dict_config
import threading


class MetlogClientHolder(object):
    """
    This is meant to be used as a singleton class that will hold references to
    MetlogClient instances and any required process-wide config data.
    """
    def __init__(self):
        self._clients = dict()
        self.global_config = dict()
        self.lock = threading.Lock()  # write lock for adding clients

    def get_client(self, name):
        """
        Return the specified MetlogClient, creating it if it doesn't exist.
        *NOTE*: Auto-created MetlogClient instances will *not* yet be usable,
        it is the downstream developer's responsibility to provide them with a
        working sender.

        :param name: String token identifying the client, also used as the
                     client's `logger` value.
        """
        client = self._clients.get(name)
        if client is None:
            with self.lock:
                # check again to make sure nobody else got the lock first
                client = self._clients.get(name)
                if client is None:
                    # TODO: there is no sender set here - grab one
                    # based on the globalconfig
                    client = MetlogClient(sender=None, logger=name)
                    if (not self._clients
                        and not self.global_config.get('default')):
                        # first one, set as default
                        self.global_config['default'] = name
                    self._clients[name] = client
        return client

    def set_client(self, name, client):
        """
        Provides a way to add a pre-existing MetlogClient to the ones stored
        in the holder.
        """
        with self.lock:
            self._clients[name] = client
            if len(self._clients) == 1:
                # first one, set as default
                self.global_config['default'] = name

    def set_default_client_name(self, name):
        """
        Convenience method for specifying what should be the default client.
        """
        self.global_config['default'] = name

    @property
    def default_client(self):
        """
        Return the default MetlogClient (as specified by the `default` value in
        the global_config dict).
        """
        default_name = self.global_config.get('default')
        if default_name is None:
            return
        return self._clients.get(default_name)

    def delete_client(self, name):
        """
        Deletes the specified client from the set of stored clients.

        :param name: Name of the client object to delete.
        """
        if name in self._clients:
            del self._clients[name]
        if self.global_config['default'] == name:
            del self.global_config['default']


CLIENT_HOLDER = MetlogClientHolder()


def get_client(name, config_dict=None):
    """
    Return client of the specified name from the CLIENT_HOLDER.

    :param name: String token to identify the MetlogClient, also used for the
                 default `logger` value of that client. `ValueError` will be
                 raised if a config is provided w/ a different `logger` value.
    :param config_dict: Configuration dictionary to be applied to the fetched
                        client.
    """
    client = CLIENT_HOLDER.get_client(name)
    if config_dict:
        logger = config_dict.get('logger')
        if logger and logger != name:
            raise ValueError('Config `logger` value must either match `name` '
                             'argument or be left blank.')
        if not logger:
            config_dict['logger'] = name
        client = client_from_dict_config(config_dict, client=client)
    return client



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2012
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Rob Miller (rmiller@mozilla.com)
#
# ***** END LICENSE BLOCK *****
from __future__ import absolute_import
try:
    import simplejson as json
except ImportError:
    import json  # NOQA

import threading
import sys
import time

if 'gevent.monkey' in sys.modules:
    from gevent import queue as Queue
else:
    import Queue  # NOQA

try:
    if 'gevent.monkey' in sys.modules:
        from gevent_zeromq import zmq
    else:
        import zmq  # NOQA
except ImportError:
    zmq = None  # NOQA


# We need to set the maximum number of outbound messages so that
# applications don't consume infinite memory if outbound messages are
# not processed
MAX_MESSAGES = 1000


class BaseClient(object):
    def __init__(self, context):
        self.context = context

        # We need to synchronize around the connected flag
        self._connect_lock = threading.RLock()
        self.set_connected(False)

    def set_connected(self, state):
        with self._connect_lock:
            self._connected = state
        return self.connected()

    def connected(self):
        with self._connect_lock:
            return self._connected


class SimpleClient(BaseClient):
    def __init__(self, context, connect_bind, hwm=200):
        super(SimpleClient, self).__init__(context)

        self.connect_bind = connect_bind
        self.hwm = hwm
        self.socket = None

        # The pub socket must be created
        # Socket to actually do pub/sub
        self.socket = self.context.socket(zmq.PUB)
        for bindstr in self.connect_bind:
            self.socket.connect(bindstr)
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.setsockopt(zmq.HWM, self.hwm)
        self.set_connected(True)

    def connect(self):
        """
        For the SimpleClient, connect() does nothing as the connect is
        handled in the initializer
        """
        return True

    def send(self, msg):
        self.socket.send(msg)


class HandshakingClient(BaseClient):
    def __init__(self, context, handshake_bind, connect_bind,
                 handshake_timeout=200,
                 hwm=200):
        super(HandshakingClient, self).__init__(context)

        self.handshake_bind = handshake_bind
        self.connect_bind = connect_bind

        self.handshake_timeout = handshake_timeout
        self.hwm = hwm

        self.handshake_socket = None
        self.socket = None

        # Socket to actually do pub/sub
        self.socket = self.context.socket(zmq.PUB)
        self.socket.connect(self.connect_bind)
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.setsockopt(zmq.HWM, self.hwm)

    def connect(self):
        """
        Connect To the 0mq REPL socket and attempt a handshake to
        ensure we're properly connected.
        """
        # Socket to send handshake signals
        self.handshake_socket = None
        try:
            self.handshake_socket = self.context.socket(zmq.REQ)
            self.handshake_socket.connect(self.handshake_bind)
            self.handshake_socket.setsockopt(zmq.LINGER, 0)

            poll = zmq.Poller()
            poll.register(self.handshake_socket, zmq.POLLIN)

            self.handshake_socket.send("")
            socks = dict(poll.poll(self.handshake_timeout))

            if socks.get(self.handshake_socket) == zmq.POLLIN:
                self.handshake_socket.recv()
                return self.set_connected(True)
            else:
                return self.set_connected(False)
        finally:
            # Shutdown the handshake
            if self.handshake_socket != None:
                self.handshake_socket.close()

    def send(self, msg):
        try:
            if self.connected():
                self.socket.send(msg)
            else:
                sys.stderr.write("%s\n" % msg)
                sys.stderr.flush()
        except zmq.ZMQError:
            sys.stderr.write("%s\n" % msg)
            sys.stderr.flush()


class Pool(object):
    """
    This is a threadsafe pool of 0mq clients.

    :param client_factory:
        a factory function that creates Client instances
    :param size:
        The number of clients to create in the pool
    :param livecheck:
        The time in seconds to wait to ping the server
        from each client
    """

    def __init__(self, client_factory, size=10, livecheck=10):
        self._stop_lock = threading.RLock()
        self._stopped = False

        self._clients = Queue.Queue()
        self._livecheck = livecheck

        # This list is only used to handle reconnects if the
        # connection to the 0mq subscriber dies
        self._all_clients = []

        for i in range(size):
            client = client_factory()
            self._clients.put(client)
            self._all_clients.append(client)

        # Connect the clients on a background thread so that we can
        # startup quickly
        self._connect_thread_started = False

        def background_thread():
            while True:
                for client in self._all_clients:
                    client.connect()
                if self.is_stopped():
                    break
                time.sleep(self._livecheck)

        self._connect_thread = threading.Thread(target=background_thread)
        self._connect_thread.daemon = True

        self.start_reconnecting()

    def start_reconnecting(self):
        """
        Start the background thread that handles pings to the server
        to synchronize the initial pub/sub
        """
        with self._stop_lock:
            if self._connect_thread_started:
                return
            self._connect_thread_started = True

        self._connect_thread.start()

    def send(self, msg):
        """
        Threadsafely send a single text message over a 0mq socket
        """
        sock = None
        try:
            sock = self.socket()
            sock.send(msg)
        except Queue.Empty:
            # Sometimes, we'll get nothing
            sys.stderr.write("%s\n" % msg)
        finally:
            if sock:
                self._clients.put(sock)

    def stop(self):
        """
        Shutdown the background reconnection thread
        """
        with self._stop_lock:
            self._stopped = True

    def is_stopped(self):
        with self._stop_lock:
            return self._stopped

    def socket(self):
        return self._clients.get()


class ZmqSender(object):
    """
    Base class for ZmqPubSender and ZmqHandshakePubSender
    """

    _zmq_context = zmq.Context() if zmq is not None else None

    def __new__(cls, *args, **kwargs):
        """
        Just check that we have pyzmq installed
        """
        if zmq is None:
            # no zmq -> no ZmqPubSender
            raise ValueError('Must have `pyzmq` installed to use ZmqPubSender')
        return super(ZmqSender, cls).__new__(cls)

    def send_message(self, msg):
        """
        Serialize and send a message off to the metlog listener.

        :param msg: Dictionary representing the message.
        """
        json_msg = json.dumps(msg)
        if self.debug_stderr:
            sys.stderr.write(json_msg + '\n')
            sys.stderr.flush()
        self.pool.send(json_msg)


class ZmqPubSender(ZmqSender):
    """
    Sends metlog messages out via a ZeroMQ publisher socket.
    """

    def __init__(self, bindstrs,
                 pool_size=10,
                 queue_length=MAX_MESSAGES,
                 livecheck=10,
                 debug_stderr=False):
        """
        :param bindstrs:
            One or more URL strings which 0mq recognizes as an
            endpoint URL. Either a string or a list of strings is
            accepted.
        :param pool_size:
            The number of connections we maintain to the 0mq backend
        :param livecheck:
            Polling interval in seconds between client.connect() calls
        :param debug_stderr:
            Boolean flag to send messages to stderr in addition to the
            actual 0mq socket
        """

        if isinstance(bindstrs, basestring):
            bindstrs = [bindstrs]

        def get_client():
            return SimpleClient(self._zmq_context,
                                bindstrs,
                                queue_length)

        self.pool = Pool(client_factory=get_client,
                size=pool_size,
                livecheck=livecheck)
        self.debug_stderr = debug_stderr


class ZmqHandshakePubSender(ZmqSender):
    """
    Sends metlog messages out via a ZeroMQ publisher socket.

    Redirect all dropped messages to stderr
    """

    def __init__(self, handshake_bind, connect_bind,
            handshake_timeout, pool_size=10, hwm=200,
            livecheck=10,
            debug_stderr=False):
        """
        :param handshake_bind:
            A single 0mq recognized endpoint URL.
            This should point to the endpoint for handshaking of
            connections
        :param connect_bind:
            A single 0mq recognized endpoint URL.
            This should point ot the endpoint for sending actual
            Metlog messages.
        :param handshake_timeout:
            Timeout in ms to wait for responses from the 0mq server on
            handshake
        :param pool_size:
            The number of connections we maintain to the 0mq backend
        :param hwm:
            High water mark. Set the maximum number of messages to
            queue before dropping messages in case of a slow reading
            0mq server.
        :param livecheck:
            Polling interval in seconds between client.connect() calls
        :param debug_stderr:
            Boolean flag to send messages to stderr in addition to the
            actual 0mq socket
        """

        def get_client():
            client = HandshakingClient(self._zmq_context,
                                handshake_bind, connect_bind,
                                handshake_timeout, hwm)
            # Try to get all clients to connect right away
            client.connect()
            return client

        self.pool = Pool(client_factory=get_client,
                size=pool_size,
                livecheck=livecheck)
        self.debug_stderr = debug_stderr
